﻿using System;
using System.IO;
using System.CodeDom.Compiler;

namespace Plywood.Compiler {
    public class CodeBuilder {
        private Class[] p_Classes = new Class[0];
        private string[] p_NamespaceIncludes = new string[0];
        private string[] p_References = new string[0];
        private Method p_EntryPoint;
        
        public CodeBuilder() {
            //Define the initial class and add it
            Class program = new Class("Program") {  
                Accessor = Accessor.Static
            };
            AddClass(program);
            
            //Define and a the main entry point
            Method entryPoint = new Method("Main", null, new Type[] { typeof(string[]) }) {
                Accessor = Accessor.Static
            };
            program.AddMethod(entryPoint);

            //Set the selected method and class to the entry point
            SelectedMethod = entryPoint;
            SelectedClass = program;
        }

        public bool ClassExists(string name) { return GetClass(name) != null; }
        public Class GetClass(string name) { 
            //Look for the class
            foreach (Class c in p_Classes) {
                if (c.Name == name) { return c; }
            }

            //Not found
            return null;
        }
        public void AddClass(Class cls) { 
            //Already exists?
            if (ClassExists(cls.Name)) {
                throw new Exception("Class already exists");
            }

            //Set the classes code builder
            cls.setBuilder(this);

            //Add it
            Array.Resize(ref p_Classes, p_Classes.Length + 1);
            p_Classes[p_Classes.Length - 1] = cls;
        }
        public void AddNetFrameworkAssembly(string name) {
            //Define the physical path
            string pPath =  new FileInfo(Environment.SystemDirectory + "../../Microsoft.NET/Framework/v2.0.50727/" + name + ".dll").FullName;

            //Already added?
            foreach (string s in p_References) {
                if (new FileInfo(s).FullName.ToLower() == new FileInfo(pPath).FullName.ToLower()) {
                    return;
                }
            }

            //Add the assembly reference
            Array.Resize(ref p_References, p_References.Length + 1);
            p_References[p_References.Length - 1] = pPath;
        }
        public void AddNamespaceInclude(string name) {
            //Already added?
            foreach (string s in p_NamespaceIncludes) {
                if (s == name) { return; }
            }

            //Add it
            Array.Resize(ref p_NamespaceIncludes, p_NamespaceIncludes.Length + 1);
            p_NamespaceIncludes[p_NamespaceIncludes.Length - 1] = name;
        }
        public void Clear() { 
            //Reset everything
            p_NamespaceIncludes = new string[0];
            p_References = new string[0];
            p_Classes = new Class[0];
            p_EntryPoint = null;
            SelectedMethod = null;
            SelectedClass = null;
        }
        public void ClearClasses() { 
            //Just restart the class array
            p_Classes = new Class[0];
            SelectedClass = null;
            SelectedMethod = null;
        }

        public void Compile(string filename) { Compile(filename, false); }
        public void Compile(string filename, bool library) {
            //Define the compiler data
            CompilerParameters param = new CompilerParameters();
            CodeDomProvider provider = Microsoft.CSharp.CSharpCodeProvider.CreateProvider("cs");
            CompilerResults results;
            ICodeCompiler compiler = provider.CreateCompiler();
            param.GenerateExecutable = !library;
            param.GenerateInMemory = false;
            foreach (string s in p_References) {
                param.ReferencedAssemblies.Add(s);
            }
              
            //Compile it
            results = compiler.CompileAssemblyFromSource(param, CSCode);

            //Any errors?
            if (results.Errors.HasErrors) {
                string error = "";
                foreach (CompilerError s in results.Errors) { error += s.ToString(); }
                error = error.Substring(0, error.Length - 1);
                throw new Exception(error);
            }

            //File exists?
            if (File.Exists(filename)) {
                File.Delete(filename);
            }

            //Copy the assembly to the filename
            File.Copy(results.CompiledAssembly.CodeBase.Replace("file:///", ""), filename);
        }

        public class Method {
            private Type[] p_ArgumentTypes;
            private Type p_ReturnType;
            private string p_Name;
            private string[] p_InlineCode = new string[0];
            private Accessor p_Accessor = Accessor.Public;
            private Class p_ParentClass;

            public Method(string name, Type returnType, Type[] argTypes) { 
                //Define the globals
                p_Name = name;
                p_ArgumentTypes = argTypes;
                p_ReturnType = returnType;
            }

            public string Name { get { return p_Name; } }
            public Type ReturnType { get { return p_ReturnType; } }
            public Type[] Arguments { get { return p_ArgumentTypes; } }
            public Accessor Accessor { get { return p_Accessor; } set { 
                //Valid?
                if (!validAccessor((int)value)) {
                    throw new Exception("Invalid accessor");
                }

                //Set the value
                p_Accessor = value;
            } }
            public Class Parent { get { return p_ParentClass; } }

            public string GetLine(long index) { return p_InlineCode[index]; }
            public long AddLine(string line) { 
                //Add the line to the inline code
                Array.Resize(ref p_InlineCode, p_InlineCode.Length + 1);
                p_InlineCode[p_InlineCode.Length - 1] = line;
                return p_InlineCode.Length - 1;
            }
            public string CSCode { get { 
                //Define the return buffer
                string buffer = "";
                
                //Add the begin of the method
                buffer += accessorStr((int)p_Accessor) + " " + DynamicLibraryLoader.getCSType(p_ReturnType) + " " + Name +
                                        "(";

                //Add the method argument code
                for (int c = 0; c < p_ArgumentTypes.Length; c++) {
                    buffer += DynamicLibraryLoader.getCSType(p_ArgumentTypes[c]) + " arg" + c.ToString() + (c == p_ArgumentTypes.Length - 1 ? "" : ",");
                }
                buffer += "){";

                //Clean up
                foreach (string s in p_InlineCode) { buffer += s; }
                buffer += "}";

                return buffer;
            } }

            internal void setParent(Class cls) { p_ParentClass = cls; }
        }
        public class Class {
            private string p_Name;
            private Accessor p_Accessor;
            private Method[] p_Methods;
            private Class[] p_Classes;
            private string[] p_InlineCode = new string[0];
            private Class p_Parent;
            private CodeBuilder p_Builder;

            public Class(string name) { 
                //Define the globals
                p_Name = name;
                p_Methods = new Method[0];
                p_Classes = new Class[0];
            }

            public string Name { get { return p_Name; } }
            public Accessor Accessor { get { return p_Accessor; } set { 
                //Valid?
                if (!validAccessor((int)value)) {
                    throw new Exception("Invalid accessor");
                }

                //Set the value
                p_Accessor = value;
            } }

            public void AddMethod(Method method) { 
                //Already exist?
                foreach (Method m in p_Methods) {
                    if (m.Name == method.Name) {
                        throw new Exception("Method already exists");
                    }
                }

                //Is this method an entry point?
                if (method.Name == "Main") {
                    if (!accessorStr((int)method.Accessor).Contains("static") || p_Builder.p_EntryPoint != null) {
                        throw new Exception("Invalid method for entry point");
                    }                    
                    p_Builder.p_EntryPoint = method;
                }

                //Set the methods parent
                method.setParent(this);

                //Add it
                Array.Resize(ref p_Methods, p_Methods.Length + 1);
                p_Methods[p_Methods.Length - 1] = method;
            }
            public void AddClass(Class cls) { 
                //Is this class the same as this?
                if (cls.Name == Name || cls == this) {
                    throw new Exception("Cannot add parent class as a child");
                }
                
                //Already exist?
                if (cls.Name == Name) { throw new Exception("Class already exists"); }
                foreach (Class c in p_Classes) {
                    if (c.Name == cls.Name) {
                        throw new Exception("Class already exists");
                    }
                }

                //Set it's parent
                cls.p_Parent = this;

                //Add it
                Array.Resize(ref p_Classes, p_Classes.Length + 1);
                p_Classes[p_Classes.Length - 1] = cls;
            }

            public Class GetClass(string name) { 
                //Find the class
                foreach (Class cls in p_Classes) {
                    if (cls.Name == name) { return cls; }
                }

                //Not found
                return null;
            }
            public Method GetMethod(string name) { 
                //Find the method
                foreach (Method m in p_Methods) {
                    if (m.Name == name) { return m; }
                }

                //Not found
                return null;
            }
            public bool ClassExists(string name) { return GetClass(name) != null; }
            public bool MethodExists(string name) { return GetMethod(name) != null; }

            public void AddDllImport(string methodName, string file, Type returnType, Type[] argumentTypes, bool privateAccess) { 
                //Get the argument code
                string argCode = "";
                for (int c = 0; c < argumentTypes.Length; c++) {
                    argCode += DynamicLibraryLoader.getCSType(argumentTypes[c]) + " dll" + c.ToString() + (c == argumentTypes.Length - 1 ? "" : ",");
                }

                //Add the namespace import
                p_Builder.AddNetFrameworkAssembly("mscorlib");
                p_Builder.AddNamespaceInclude("System.Runtime.InteropServices");
                
                //Define the return buffer
                string buffer = "";

                //Add the DllImport attribute information
                buffer += "[DllImport(\"" + file + "\")]";

                //Get the code
                buffer += (privateAccess ? "private" : "public") + " static extern " + DynamicLibraryLoader.getCSType(returnType) + " " + methodName;
                buffer += "(" + argCode + ");";

                //Add the code
                AddLine(buffer);
            }

            public string GetLine(long index) { return p_InlineCode[index]; }
            public long AddLine(string line) { 
                //Add the line to the inline code
                Array.Resize(ref p_InlineCode, p_InlineCode.Length + 1);
                p_InlineCode[p_InlineCode.Length - 1] = line;
                return p_InlineCode.Length - 1;
            }
            public string CSCode { get {
                //Define the return buffer
                string buffer = "";

                //Define the initial class
                buffer += accessorStr((int)p_Accessor) + " class " + Name + "{";

                //Add the inline code
                foreach (string s in p_InlineCode) { buffer += s; }

                //Add the method code
                foreach (Method m in p_Methods) {
                    buffer += m.CSCode;
                }

                //Add the classes code
                foreach (Class c in p_Classes) {
                    buffer += c.CSCode;
                }

                //Return the buffer
                return buffer + "}";
            } }
            public Class Parent { get { return p_Parent; } }

            public void Clear() { 
                //Is one of the methods a selected method?
                foreach (Method m in p_Methods) {
                    if (m == p_Builder.SelectedMethod) { p_Builder.SelectedMethod = null; break; }
                }

                //Reset the method array
                p_Methods = new Method[0];
            }

            internal void setBuilder(CodeBuilder b) { p_Builder = b; }
        }

        public string CSCode {
            get {
                //Define the return buffer
                string buffer = "";

                //Add namespace includes
                foreach (string s in p_NamespaceIncludes) {
                    buffer += "using " + s + ";";
                }

                //Add the classes
                foreach (Class c in p_Classes) {
                    buffer += c.CSCode;
                }

                //Return the buffer
                return buffer;
            }
        }
        public Method EntryPoint { get { return p_EntryPoint; } }
        public Method SelectedMethod { get; set; }
        public Class SelectedClass { get; set; }

        private static bool validAccessor(int accessor) { 
            //All 3?
            if (accessor == (int)(Accessor.Public | Accessor.Private | Accessor.Static)) { return false; }
            //Public and private?
            if (accessor == (int)(Accessor.Public | Accessor.Private)) { return false; }

            //Valid 
            return true;
        }
        private static string accessorStr(int accessor) { 
            //Only static?
            if (accessor == (int)Accessor.Static) { return "static"; }
            
            //Private/public only?
            if (accessor == (int)Accessor.Public || accessor == (int)Accessor.Private) {
                return (accessor == (int)Accessor.Public) ? "public" : "private";
            }

            //Return private/public static
            if (accessor == (int)(Accessor.Public | Accessor.Static)) { return "public static"; }
            return "private static";
        }
                
        public enum Accessor { 
            Public = 1,
            Static = 2, 
            Private = 4,
        }

    }
}