﻿using System;
using System.Windows.Forms;
using System.Drawing;
using System.IO;
using System.Drawing.Drawing2D;

namespace Plywood {
    public class DevelopmentEnvironment {
        private bool p_Shown;
        private RichTextBox p_Code;
        private Form p_Form;
        private bool p_IsUntitled = true;
        private string p_File;
        

        #region UI
        public Form CreateForm() { 
            //Already shown?
            if (p_Shown) { throw new Exception("Only one form can be shown"); }
            p_Shown = true;

            //Define the form
            Form form = new Form();
            p_Form = form;

            //Define the form globals
            form.Text = "Untitled - " + baseText;
            form.ShowIcon = false;
            form.Size = new Size(400, 300);
            form.FormClosed += delegate(object sender, FormClosedEventArgs e) { p_Shown = false; };

            //Add the menu
            MenuStrip menu = new MenuStrip();
            menu.ForeColor = Color.White;
            addFileDropDown((ToolStripMenuItem)menu.Items.Add("File"));
            addEditDropDown((ToolStripMenuItem)menu.Items.Add("Edit"));
            addToolsDropDown((ToolStripMenuItem)menu.Items.Add("Tools"));
            form.Controls.Add(menu);
            Bitmap bit = new Bitmap(1, menu.Height);
            Graphics.FromImage(bit).FillRectangle(
                    new LinearGradientBrush(new Point(0, 0), new Point(0, menu.Height),
                                            Color.FromArgb(255, 30, 30, 30),
                                            Color.Black),
                    new Rectangle(0, 0, 1, menu.Height));
            menu.BackgroundImage = bit;

            //Add the code control
            p_Code = new RichTextBox();
            p_Code.Dock = DockStyle.Fill;
            p_Code.BackColor = Color.Black;
            p_Code.Font = new Font("Arial", 8, FontStyle.Bold);
            p_Code.ForeColor = Color.White;
            p_Code.BorderStyle = BorderStyle.None;            
            form.Controls.Add(p_Code);
            p_Code.BringToFront();
            return form;
        }
        
        private void showList(string[] list, string text) { 
            //Define the form
            Form form = new Form();
            form.Text = text + " listing: " + list.Length.ToString() + " entr" + (list.Length != 1 ? "ies" : "y");
            form.ShowInTaskbar = false;
            form.ShowIcon = false;
            form.StartPosition = FormStartPosition.CenterParent;

            //Create the list control
            ListBox lb = new ListBox();
            foreach (string s in list) {
                lb.Items.Add(s);
            }
            lb.Dock = DockStyle.Fill;
            form.Controls.Add(lb);

            //Show it
            form.Show();
        }
        private void saveAsClick(object sender, EventArgs e) { 
            //Save it
            SaveFileDialog save = new SaveFileDialog();
            save.Filter = filter;
            if (save.ShowDialog() == DialogResult.OK) {
                File.WriteAllText(save.FileName, p_Code.Text);
                p_File = save.FileName;
                p_IsUntitled = false;
                p_Form.Text = new FileInfo(save.FileName).Name + " - " + baseText;
            }
        }
        private void saveClick(object sender, EventArgs e) { 
            //Already saved?
            if (p_IsUntitled) { saveAsClick(sender, e); return; }

            //Save
            File.WriteAllText(p_File, p_Code.Text);
        }
        private void openClick(object sender, EventArgs e) {
            //Open a file
            OpenFileDialog opn = new OpenFileDialog();
            opn.Filter = filter;
            if (opn.ShowDialog() == DialogResult.OK) {
                p_IsUntitled = false;
                p_Code.Text = File.ReadAllText(opn.FileName);
                p_File = opn.FileName;
                p_Form.Text = new FileInfo(opn.FileName).Name + " - " + baseText;
            }
        }
        private void exitClick(object sender, EventArgs e) {
            //Close the form
            p_Form.Close();
        }
        private void runClick(object sender, EventArgs e) { 
            //Run the code
            Interpreter i = new Interpreter();
            i.Run(i.createStrStream(p_Code.Text));
        }
        private void listMethodCallsClick(object sender, EventArgs e) {
            //Get methods
            Interpreter i = new Interpreter();
            IMethod[] method = null;
            string[] name = null;
            i.ListMethodCalls(i.createStrStream(p_Code.Text), out method, out name);

            //Define the string to display
            string display = "";
            foreach (IMethod m in method) {
                string l = m.Name + "(";
                foreach (Type t in m.ArgumentTypes) { l += t.FullName + ", "; }
                if (l.Length != 0) { l = l.Substring(0, l.Length - 2); }
                display += l + ")\n";
            }
            if (display.Length != 0) { display = display.Substring(0, display.Length - 1); }

            //Display it
            showList(display.Split('\n'), "Method calls");
        }
        private void listPlywoodMethodsClick(object sender, EventArgs e) { 
            //Show the plywood methods
            IMethod[] methods = new GenericMethods(null).GetClrMethods();
            string[] list = new string[methods.Length];
            for (int c = 0; c < methods.Length; c++) {
                string line = methods[c].Name + " (";
                foreach (Type t in methods[c].ArgumentTypes) {
                    line += t.FullName + ", ";
                }
                if (methods[c].ArgumentTypes.Length != 0) {
                    line = line.Substring(0, line.Length - 2);
                }
                line += ")";
                list[c] = line;
            }

            //Show the list
            showList(list, "Generic methods list");
        }
        private void addFileDropDown(ToolStripMenuItem item) {
            //Add the buttons
            item.DropDown.Items.Add("Save", null, saveClick);
            item.DropDown.Items.Add("Save as", null, saveAsClick);
            item.DropDown.Items.Add(new ToolStripSeparator());
            item.DropDown.Items.Add("Open", null, openClick);
            item.DropDown.Items.Add(new ToolStripSeparator());
            item.DropDown.Items.Add("Exit", null, exitClick);

            //Set the shortcut keys
            ((ToolStripMenuItem)item.DropDown.Items[0]).ShortcutKeys = Keys.Control | Keys.S;
            ((ToolStripMenuItem)item.DropDown.Items[1]).ShortcutKeys = Keys.Control | Keys.Shift | Keys.S;
            ((ToolStripMenuItem)item.DropDown.Items[3]).ShortcutKeys = Keys.Control | Keys.O;
            ((ToolStripMenuItem)item.DropDown.Items[5]).ShortcutKeys = Keys.Alt | Keys.F4;
        }
        private void addToolsDropDown(ToolStripMenuItem item) { 
            //Add the items
            item.DropDown.Items.Add("Run", null, runClick);
            item.DropDown.Items.Add("List script method calls", null, listMethodCallsClick);
            item.DropDown.Items.Add("List generic methods", null, listPlywoodMethodsClick);

            //Set the shortcut keys
            ((ToolStripMenuItem)item.DropDown.Items[0]).ShortcutKeys = Keys.F5;

        }
        private void addEditDropDown(ToolStripMenuItem item) {
            //Add the items
            item.DropDown.Items.Add("Cut", null, cutClick);
            item.DropDown.Items.Add("Copy", null, copyClick);
            item.DropDown.Items.Add("Paste", null, pasteClick);
            item.DropDown.Items.Add(new ToolStripSeparator());
            item.DropDown.Items.Add("Undo", null, undoClick);
            item.DropDown.Items.Add("Redo", null, redoClick);
            item.DropDown.Items.Add(new ToolStripSeparator());
            item.DropDown.Items.Add("Select all", null, selectAllClick);

            //Set the shortcut keys
            ((ToolStripMenuItem)item.DropDown.Items[0]).ShortcutKeys = Keys.Control | Keys.X;
            ((ToolStripMenuItem)item.DropDown.Items[1]).ShortcutKeys = Keys.Control | Keys.C;
            ((ToolStripMenuItem)item.DropDown.Items[2]).ShortcutKeys = Keys.Control | Keys.V;
            ((ToolStripMenuItem)item.DropDown.Items[4]).ShortcutKeys = Keys.Control | Keys.Z;

        }
        private void copyClick(object sender, EventArgs e) { 
            //Copy
            p_Code.Copy();
        }
        private void cutClick(object sender, EventArgs e) { 
            //Cut
            p_Code.Cut();
        }
        private void pasteClick(object sender, EventArgs e) { 
            //Paste
            p_Code.Paste();
        }
        private void undoClick(object sender, EventArgs e) { 
            //Undo
            p_Code.Undo();
        }
        private void redoClick(object sender, EventArgs e) { 
            //Redo
            p_Code.Redo();
        }
        private void selectAllClick(object sender, EventArgs e) { 
            //Select all
            p_Code.SelectAll();
        }

        private string baseText { get { return "Plywood Script IDE"; } }
        private string filter { get { return "Plywood Source File|*.ply"; } }
        #endregion

    }
}