﻿using System;
using System.IO;
using Microsoft.Win32;
using System.IO.Compression;

namespace Plywood {
    public class FileSystem {
        private Stream p_Stream;
        private bool p_Hang;
        private byte[] p_MagicKey = null;
        private RegistryKey p_Registry;
        private RegistryKey p_AliasList;
       
        public FileSystem(Stream stream) { _constructor(stream); }
        public FileSystem(string filename) { 
            //Open a file stream
            _constructor(new FileStream(filename, FileMode.OpenOrCreate));
        }
        
        private void _constructor(Stream str) {
            //Check the stream
            if (!str.CanWrite || !str.CanSeek || !str.CanRead) { 
                throw new Exception("Stream must be able to seek, read and write");
            }

            //Define the entry magic key
            p_MagicKey = new byte[] { 
                (byte)'F',
                (byte)'D',
                (byte)'E', 230, 1, 22
            };
            
            //Check the date
            if (DateTime.Now.Ticks < baseTime.Ticks) {
                throw new Exception("Invalid time and date settings");
            }

            //Define the globals
            p_Stream = str;
            BufferSize = 10 * 1024 * 1024;
            FileEntryArgumentLength = 6;
            p_Registry = Registry.CurrentUser.CreateSubKey("SOFTWARE\\Social\\Candle\\Update\\FileSystem");
        
            //Create the file system alias registry
            p_AliasList = p_Registry.CreateSubKey("AliasRef");
        }

        private static string temporaryFilename() { 
            //Loop until the filename is vacant
            while (true) { 
                //Generate a filename
                string filename = "";
                Random r = new Random();
                for (int c = 0; c < 25; c++) {
                    char ch = (char)r.Next((int)'A', (int)'Z');
                    if (r.Next(0, 5) == 1) { ch = ch.ToString().ToLower()[0]; }
                    filename += ch;
                }

                //Add the extension
                filename += ".tmp";

                //Add the directory
                filename = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData) + "\\" + filename;
                
                //Is this file available?
                if (!File.Exists(filename)) { 
                    //Return the filename
                    return filename;
                }
            }
        }
        private DateTime getDateTime(long ticks) { 
            //Add the ticks to the base time and return it
            return baseTime.AddTicks(ticks);
        }
        private long getDateTime(DateTime time) { 
            //Subtract the ticks from this time by the base time
            return time.Ticks - baseTime.Ticks;
        }
        internal static string createRef(long start, long length) { 
            //Check the length
            if (length > 10737418240) { //10gb
                throw new Exception("Length cannot exceed 10gb (10737418240 bytes)");
            }

            //Check the positioning
            if (start < 0 || length < 0) {
                throw new Exception("Invalid reference");
            }

            //Grab the start and length as hex
            string sHex = NumericOperations.ToHex(NumericOperations.ToBinary((ulong)start));
            string sLength = NumericOperations.ToHex(NumericOperations.ToBinary((ulong)length));

            //Force hex length of 16 for the start index
            while (sHex.Length < 16) {
                sHex = "0" + sHex;
            }

            //Force the hex length of 9 for the length
            while (sLength.Length < 9) {
                sLength = "0" + sLength;
            }

            //Return the start and length hex values together
            return sHex + sLength;
        }
        private byte[] getHexBytes(string hex) { 
            //Force hex as a multiple of 2
            while ((hex.Length * 1.0f / 2).ToString().Split('.').Length != 1) {
                hex = "0" + hex;
            }

            //Define the return buffer
            byte[] buffer = new byte[hex.Length / 2];

            //Fill in the hex values
            for (int c = 0; c < buffer.Length; c++) { 
                //Grab the hex in the value
                string h = hex.Substring(c * 2, 2);

                //Add the decimal value for the hex
                buffer[c] = (byte)NumericOperations.FromBinary(NumericOperations.FromHex(h));
            }

            //Return the buffer
            return buffer;
        }
        private string bytesToHex(byte[] bytes) { 
            //Define the return buffer
            string buffer = "";

            //Add all the hex values
            foreach (byte b in bytes) {
                string hex = NumericOperations.ToHex(NumericOperations.ToBinary((ulong)b));
                while (hex.Length < 2) {
                    hex = "0" + hex;
                }
                buffer += hex;
            }

            //Return the buffer
            return buffer;
        }
        internal void parseRef(ref string reference, out long start, out long length) { 
            //Let the local method handle this
            parseRef(ref reference, out start, out length, true);
        }
        internal void parseRef(ref string reference, out long start, out long length, bool checkAlias) {
            //Is the reference a alias?
            if (p_AliasList.GetValue(reference) != null && checkAlias) {
                reference = p_AliasList.GetValue(reference).ToString();
            }

            //Check the reference length
            if (reference.Length != 25)  {
                throw new Exception("Invalid reference");
            }

            //Get the start and lengh
            start = (long)NumericOperations.FromBinary(NumericOperations.FromHex(reference.Substring(0, 16)));
            length = (long)NumericOperations.FromBinary(NumericOperations.FromHex(reference.Substring(16, 9)));
        }
        private byte[] getFileArguments(DateTime created, DateTime accessed, byte[] attributes) { 
            //Get the hex value for the write times
            string createHex = NumericOperations.ToHex(NumericOperations.ToBinary((ulong)getDateTime(created)));
            string accessHex = NumericOperations.ToHex(NumericOperations.ToBinary((ulong)getDateTime(accessed)));

            //Force hex lengths of 16 for the date time values
            while (createHex.Length < 16) { createHex = "0" + createHex; }
            while (accessHex.Length < 16) { accessHex = "0" + accessHex; }

            //Get the byte values for the hex values
            byte[] createBytes = getHexBytes(createHex);
            byte[] accessBytes = getHexBytes(accessHex);

            //Define the return buffer
            byte[] buffer = new byte[createBytes.Length + accessBytes.Length + attributes.Length];

            //Set the values
            for (int c = 0; c < createBytes.Length; c++) {
                buffer[c] = createBytes[c];
            }
            for (int c = 0; c < accessBytes.Length; c++) {
                buffer[c + createBytes.Length] = accessBytes[c];
            }
            for (int c = 0; c < attributes.Length; c++) {
                buffer[c + createBytes.Length + accessBytes.Length] = attributes[c];
            }

            //Return the buffer
            return buffer;
        }

        public AsyncFileSystemIO BeginCreateFile(byte[] attributes) { 
            //Check argument count
            if (attributes.Length != FileEntryArgumentLength) {
                throw new Exception("Invalid arguments");
            }

            //Hang?
            while (p_Hang) ;
                     
            //Cause hang
            p_Hang = true;

            //Seek to the end of the stream
            p_Stream.Seek(0, SeekOrigin.End);

            //Write the file information
            byte[] info = getFileArguments(DateTime.Now, DateTime.Now, attributes);
            p_Stream.Write(p_MagicKey, 0, p_MagicKey.Length);
            p_Stream.Write(info, 0, info.Length);

            //Define the underlying stream
            Stream under = new _directStream { Stream = p_Stream, FileSystem = this };
            
            //Define the writer stream
            IO.Output output = new IO.Output(under);
            
            //Return the syncronous stream
            return new AsyncFileSystemIO(output, AsyncFileSystemIO.IOType.CreateFile);
        }
        public string EndCreateFile(AsyncFileSystemIO async) {
            //Check the class
            if (async.Type != AsyncFileSystemIO.IOType.CreateFile) {
                throw new Exception("Invalid type");
            }

            //Check for hang
            if (!p_Hang) {
                throw new Exception("Asyncronous operations where not begun");
            }

            //Stop hang
            p_Hang = false;

            //Grab the stream length
            long length = async.WriteStream.Length;

            //Close the stream
            async.WriteStream.BaseStream.Close();

            //Return the reference to this file
            return createRef(p_Stream.Position - length, length);
        }

        public string Compress(Stream input, byte[] attributes) {
            //Seek to the beginning of the file
            input.Position = 0;

            //Create a temporary file to compress to
            string tmpFile = temporaryFilename();

            //Compress the data
            FileStream fs = new FileStream(tmpFile, FileMode.Create);
            GZipStream cmp = new GZipStream(fs, CompressionMode.Compress);                
            CopyStream(input, cmp, BufferSize);            

            //Write the contents of the file to a new record
            cmp.Close();
            fs = new FileStream(tmpFile, FileMode.Open);
            string buffer = WriteAllStream(fs, attributes);

            //Clean up
            fs.Close();
            File.Delete(tmpFile);
            return buffer;
        }
        public string Compress(Stream input) { 
            //Let the local method handle this
            return Compress(input, new byte[FileEntryArgumentLength]);
        }
        public string Compress(byte[] input, byte[] attributes) { 
            //Write the data to a memory stream
            MemoryStream mem = new MemoryStream();
            mem.Write(input, 0, input.Length);

            //Compress the data
            mem.Position = 0;
            string buffer = Compress(mem, attributes);

            //Clean up
            mem.Close();
            return buffer;
        }
        public string Compress(byte[] input) { 
            //Let the local method handle this
            return Compress(input, new byte[FileEntryArgumentLength]);
        }
        public string Compress(string filename, byte[] attributes) { 
            //Create a file stream
            FileStream fs = new FileStream(filename, FileMode.Open);

            //Compress the data
            string buffer = Compress(fs, attributes);

            //Clean up
            fs.Close();
            return buffer;
        }
        public string Compress(string filename) { 
            //Let the local method handle this
            return Compress(filename, new byte[FileEntryArgumentLength]);
        }
        public void Decompress(string reference, Stream stream) { 
            //Create the stream for the reference
            Stream source = CreateStream(reference);

            //Decompress the stream
            Decompress(source, stream, BufferSize);

            //Clean up 
            source.Close();
        }
        public byte[] Decompress(string reference) { 
            //Create a memory stream to read to
            MemoryStream mem = new MemoryStream();

            //Decompress
            Decompress(reference, mem);

            //Write the data to the return buffer
            byte[] buffer = new byte[mem.Length];
            mem.Position = 0;
            mem.Read(buffer, 0, buffer.Length);

            //Clean up
            mem.Close();
            return buffer;
        }
        public void Decompress(string reference, string outputFile) { 
            //Create a file stream
            FileStream fs = new FileStream(outputFile, FileMode.Create);

            //Decompress to the file stream
            Decompress(reference, fs);

            //Clean up
            fs.Close();
        }

        public Stream CreateStream(string reference) { 
            //Check for the file existance (with changing it's access time as well)
            try {
                DateTime created, accessed;
                byte[] args = null;
                GetFileInfo(reference, out args, out created, out accessed, false);
            }
            catch {
                throw new Exception("Reference does not exist");
            }

            //Get the positioning of this file
            long start = 0, length = 0;
            parseRef(ref reference, out start, out length);

            //Return the stream for this file
            return new _entryStream() {
                FileStart = start, 
                FileLength = length,
                FileSystem = this
            };
        }

        public string WriteAllStream(Stream stream, byte[] attributes) { 
            //Create asyncronized write
            AsyncFileSystemIO async = BeginCreateFile(attributes);

            //Seek to the beginning of the stream
            stream.Seek(0, SeekOrigin.Begin);

            //Stream the stream data to the write stream
            for (long c = 0; c < stream.Length; c++) {
                int packetSize = BufferSize;
                if (c + packetSize > stream.Length) { packetSize = (int)(stream.Length - c); }
                byte[] packet = new byte[packetSize];
                stream.Read(packet, 0, packet.Length);
                async.WriteStream.Write(packet, 0, packet.Length);
                packet = null;
                c = stream.Position + 1;
            }

            //Return the new reference
            return EndCreateFile(async);
        }
        public string WriteAllStream(Stream stream) {
            //Let the local method handle this
            return WriteAllStream(stream, new byte[FileEntryArgumentLength]);
        }
        public string WriteAllBytes(byte[] bytes, byte[] attributes) { 
            //Define the memory stream
            MemoryStream ms = new MemoryStream();

            //Write the value
            ms.Write(bytes, 0, bytes.Length);

            //Finish writing
            string buffer = WriteAllStream(ms, attributes);
            ms.Close();
            return buffer;
        }
        public string WriteAllBytes(byte[] bytes) { 
            //Let the local method handle this
            return WriteAllBytes(bytes, new byte[FileEntryArgumentLength]);
        }
        public string WriteAllLines(string[] lines, byte[] attributes) { 
            //Rebuild the lines as a single string
            string buffer = "";
            for(int c = 0; c < lines.Length;c++) {
                buffer += lines[c] + (c == lines.Length - 1 ? "" : Environment.NewLine);
            }

            //Write the file
            return WriteAllText(buffer, attributes);
        }
        public string WriteAllLines(string[] lines) { 
            //Let the local method handle this
            return WriteAllLines(lines, new byte[FileEntryArgumentLength]);
        }
        public string WriteAllText(string text, byte[] attributes) { 
            //Let the local method handle this
            return WriteAllBytes(System.Text.Encoding.ASCII.GetBytes(text), attributes);
        }
        public string WriteAllText(string text) { 
            //Let the local method handle this
            return WriteAllText(text, new byte[FileEntryArgumentLength]);
        }

        public byte[] ReadAllBytes(string reference) { 
            //Create a stream
            Stream str = CreateStream(reference);

            //Define the return buffer
            byte[] buffer = new byte[str.Length];

            //Read the stream into the buffer
            str.Read(buffer, 0, buffer.Length);

            //Clean up
            str.Close();
            return buffer;
        }
        public string[] ReadAllLines(string reference) { 
            //Split up the text of the file into lines
            string[] lines = ReadAllText(reference).Split('\n');

            //Remove unwanted characters
            for (int c = 0; c < lines.Length; c++) {
                while (lines[c].EndsWith("\n")) {
                    try { lines[c] = lines[c].Substring(0, lines[c].Length - 1); }
                    catch { }
                }
            }

            //Return the lines
            return lines;
        }
        public string ReadAllText(string reference) { 
            //Return the string version of the read bytes
            return System.Text.Encoding.ASCII.GetString(ReadAllBytes(reference));
        }
        public void ReadAllStream(string reference, Stream stream) { 
            //Create a stream to read from
            Stream from = CreateStream(reference);
            
            //Stream the file bytes to the dest stream
            for (long c = 0; c < from.Length; c++) {
                int packetSize = BufferSize;
                if (c + packetSize > from.Length) { packetSize = (int)(from.Length - c); }
                byte[] packet = new byte[packetSize];
                from.Read(packet, 0, packet.Length);
                stream.Write(packet, 0, packet.Length);
                packet = null;
                c = from.Position + 1;
            }
        }

        public void ChangeFile(string reference, byte[] arguments) {
            //Hang?
            while (p_Hang) ;
            p_Hang = true;

            //Get the current reference information
            DateTime created, accessed;
            byte[] args = null;
            GetFileInfo(reference, out args, out created, out accessed, false);

            //Get the position of the reference
            long start, length;
            parseRef(ref reference, out start, out length);

            //Define the new file entry information
            byte[] newInfo = getFileArguments(created, DateTime.Now, arguments);

            //Write the new information
            p_Stream.Position = start - newInfo.Length;
            p_Stream.Write(newInfo, 0, newInfo.Length);

            //Stop hang
            p_Hang = false;
        }

        public bool FileExists(string reference) {
            //Attempt to get its file info
            byte[] arg = null;
            DateTime created = DateTime.MinValue;
            DateTime accessed = DateTime.MinValue;
            return GetFileInfo(reference, out arg, out created, out accessed);
        }

        public bool GetFileInfo(string reference, out byte[] arguments, out DateTime createdTime, out DateTime accessedTime) { 
            //Let the local method handle this
            return GetFileInfo(reference, out arguments, out createdTime, out accessedTime, true);
        }
        public bool GetFileInfo(string reference, out byte[] arguments, out DateTime createdTime, out DateTime accessedTime, bool preserveAccessTime) { 
            //Set default outputs
            createdTime = DateTime.MinValue;
            accessedTime = DateTime.MaxValue;
            arguments = null;
           
            //Check the reference
            long start, length;
            try { parseRef(ref reference, out start, out length); }
            catch {
                return false;
            }
            
            //Grab the new file information data (ignore created time and arguments for now)
            byte[] newFileInfo = getFileArguments(baseTime, DateTime.Now, new byte[FileEntryArgumentLength]);

            //Check the reference
            if (start < 0 || length < 0 || start - newFileInfo.Length - p_MagicKey.Length < 0 ||
                                          start + length > p_Stream.Length) {
                                              return false;
            }

            //Read the magic token
            byte[] token = new byte[p_MagicKey.Length];
            p_Stream.Position = start - newFileInfo.Length - p_MagicKey.Length;
            p_Stream.Read(token, 0, token.Length);

            //Check the token
            for (int c = 0; c < token.Length; c++) {
                if (token[c] != p_MagicKey[c]) {
                    return false;
                }
            }

            //Read the information raw bytes
            byte[] createdRaw = new byte[8];
            byte[] accessedRaw = new byte[8];
            p_Stream.Read(createdRaw, 0, 8);
            p_Stream.Read(accessedRaw, 0, 8);
            arguments = new byte[FileEntryArgumentLength];            
            p_Stream.Read(arguments, 0, arguments.Length);

            //Set the created and access time outputs            
            createdTime = getDateTime((long)NumericOperations.FromBinary(NumericOperations.FromHex(bytesToHex(createdRaw))));
            accessedTime = preserveAccessTime ?
                getDateTime((long)NumericOperations.FromBinary(NumericOperations.FromHex(bytesToHex(accessedRaw)))) :
                DateTime.Now;

            //Change the access time?
            if (!preserveAccessTime) { 
                //Grab the new file information
                newFileInfo = getFileArguments(createdTime, DateTime.Now, arguments);

                //Write the new data
                p_Stream.Position = start - newFileInfo.Length;
                p_Stream.Write(newFileInfo, 0, newFileInfo.Length);
            }

            //Succeeded
            return true;
        }

        public void CreateAlias(string reference, string alias) { 
            //Check the reference
            if (!FileExists(reference)) {
                throw new Exception("Reference does not exist");
            }

            //Does the alias already exist?
            if (p_AliasList.GetValue(alias) != null) {
                throw new Exception("Alias already exists");
            }

            //Create the alias in registry
            p_AliasList.SetValue(alias, reference);
        }
        public void RemoveAlias(string alias) { 
            //Does the alias exist?
            if (p_AliasList.GetValue(alias) == null) {
                throw new Exception("Alias does not exist");
            }

            //Remove the value
            p_AliasList.DeleteValue(alias);
        }
        public AliasReference[] GetAliasList(bool checkReference) { 
            //Define the return buffer
            AliasReference[] buffer = new AliasReference[0];

            //Loop through every alias value
            foreach (string s in p_AliasList.GetValueNames()) { 
                //Grab the value
                object val = p_AliasList.GetValue(s);

                //Check it's type
                if (!(val is string)) { continue; }

                //Check reference?
                if (checkReference && !FileExists((string)val)) { continue; }

                //Add the value
                Array.Resize(ref buffer, buffer.Length + 1);
                buffer[buffer.Length - 1] = new AliasReference(s, (string)val);
            }

            //Return the buffer
            return buffer;
        }

        private DateTime baseTime {
            get
            { 
            //Return 1st Jan 2012 to be a base value
            return new DateTime(2012, 2, 20);
        } }

        public void Close() { 
            //Close the stream
            p_Stream.Close();
        }

        public long Length { get { return p_Stream.Length; } }
        public Stream Stream { get { return p_Stream; } }
        public int BufferSize { get; set; }
        public int FileEntryArgumentLength { get; set; }

        public static void CopyStream(Stream source, Stream target, int bufferSize) { 
            //Check the streams
            if (!source.CanRead || !source.CanSeek || !target.CanWrite) {
                throw new Exception("Invalid stream, make sure that the appropriate stream has read or write capabilities.\nSource must also be seekable");
            }

            //Seek the source for writing
            source.Seek(0, SeekOrigin.Begin);

            //Stream the data from source to target
            for (long c = 0; c < source.Length; c++) {
                int packetSize = bufferSize;
                if (c + packetSize > source.Length) { packetSize = (int)(source.Length - c); }
                byte[] packet = new byte[packetSize];
                source.Read(packet, 0, packet.Length);
                target.Write(packet, 0, packet.Length);
                packet = null;
                c = source.Position + 1;
            }
        }
        public static Stream CreateTempStream(out string filename) { 
            //Generate the filename
            filename = temporaryFilename();

            //Return the file stream for this file
            return new FileStream(filename, FileMode.Create);
        }
        public static void Decompress(Stream source, Stream target, int bufferSize) { 
            //Define the decompression stream
            GZipStream compression = new GZipStream(source, CompressionMode.Decompress);

            //Loop infinity
            while (true) { 
                //Read from the compression
                byte[] packet = new byte[bufferSize];
                int r = compression.Read(packet, 0, packet.Length);

                //Write to the target stream
                target.Write(packet, 0, r);

                //Clean up
                packet = null;
                if (r == 0) { break; }
            }
            
            //Clean up
            compression.Close();
        }

        private class _entryStream : Stream {
            public FileSystem FileSystem;
            public long FileStart;
            public long FileLength;

            public override void Write(byte[] buffer, int offset, int count) {
                //Hang?
                while (FileSystem.p_Hang) ;
                FileSystem.p_Hang = true;

                //Check for overflow
                if (Position + count > FileLength) {
                    throw new Exception("Overflow");
                }

                //Encrypt the buffer?
                if (FileSystem.EncryptionKey != null && Encryption.ValidKey(FileSystem.EncryptionKey)) { 
                    Encryption.Encrypt(ref buffer, FileSystem.EncryptionKey);
                }

                //Seek the main stream for writing
                FileSystem.p_Stream.Position = FileStart + Position;

                //Write the buffer to the stream
                FileSystem.p_Stream.Write(buffer, offset, count);

                //Stop hang
                FileSystem.p_Hang = false;

                //Incriment the position
                Position += count;
            }
            public override void WriteByte(byte value) {
                //Write 1 byte
                Write(new byte[] { value }, 0, 1);
            }
            public override void SetLength(long value) {
                throw new NotSupportedException();
            }

            public override int Read(byte[] buffer, int offset, int count) {
                //Define the number of bytes to read
                int read = 0;

                //Does the position exceed?
                if (Position + count > Length) {
                    //Offset the count
                    count = (int)(Length - Position);

                    //Double check
                    if (count < 0) { throw new Exception("Invalid position"); }
                }

                //Hang?
                while (FileSystem.p_Hang) ;
                FileSystem.p_Hang = true;

                //Seek the stream to the appropriate location for reading
                FileSystem.p_Stream.Position = FileStart + Position;

                //Read the selected bytes into the buffer
                read = FileSystem.p_Stream.Read(buffer, offset, count);

                //Decrypt the buffer?
                if (FileSystem.EncryptionKey != null && Encryption.ValidKey(FileSystem.EncryptionKey)) {
                    Encryption.Decrypt(ref buffer, FileSystem.EncryptionKey);
                }

                //Set the new position
                Position += read;

                //Stop hang
                FileSystem.p_Hang = false;

                //Return the number of bytes read
                return read;
            }
            public override int ReadByte() {
                //Define the buffer to read to 
                byte[] buffer = new byte[1];
                
                //Read
                Read(buffer, 0, buffer.Length);

                //Return the first entry in the buffer
                return buffer[0];
            }

            public override long Seek(long offset, SeekOrigin origin) {
                //Define the new location
                long newLoc = 0;
                switch (origin) {
                    case SeekOrigin.Begin: newLoc = offset; break;
                    case SeekOrigin.Current: newLoc = Position + offset; break;
                    case SeekOrigin.End: newLoc = FileLength - offset; break;
                }

                //Check the position
                if (newLoc > FileStart + FileLength || newLoc < 0) {
                    throw new Exception("Out of range");
                }

                //Set the position and return the position set
                Position = newLoc;
                return newLoc;
            }
            public override void Flush() {
                throw new NotSupportedException();
            }

            public override long Length { get { return FileLength; } }
            public override long Position { get; set; }
            public override bool CanRead { get { return true; } }
            public override bool CanWrite { get { return true; } }
            public override bool CanSeek { get { return true; } }
        }
        private class _directStream : Stream {
            public Stream Stream;
            public FileSystem FileSystem;
            private long p_Position;
            private long p_Length;

            public override bool CanRead { get { return false; } }
            public override bool CanWrite { get { return true; } }
            public override bool CanSeek { get { return false; } }

            public override void SetLength(long value) {
                throw new NotSupportedException();
            }
            public override void Flush() {
                throw new NotSupportedException();
            }
            public override long Seek(long offset, SeekOrigin origin) {
                throw new NotSupportedException();
            }
            public override void Write(byte[] buffer, int offset, int count) {
                //Encrypt the buffer
                if (FileSystem.EncryptionKey != null && Encryption.ValidKey(FileSystem.EncryptionKey)) {
                    Encryption.Encrypt(ref buffer, FileSystem.EncryptionKey);
                }
                
                //Write to the stream
                Stream.Write(buffer, offset, count);
                                
                //Incriment the position
                p_Position += count;
                p_Length += count;
            }
            public override void WriteByte(byte value) {
                //Write 1 byte
                Write(new byte[] { value }, 0, 1);
            }
            public override int Read(byte[] buffer, int offset, int count) {
                throw new NotSupportedException();
            }
            public override int ReadByte() {
                throw new NotSupportedException();
            }
            public override long Position { get { return p_Position; } set { throw new NotSupportedException(); } }
            public override long Length { get { return p_Length; } }
        }

        public struct AliasReference {
            private string p_Alias, p_Reference;

            public AliasReference(string alias, string reference) {
                //Define the globals
                p_Alias = alias;
                p_Reference = reference;
            }

            public string Alias { get { return p_Alias; } }
            public string Reference { get { return p_Reference; } }
        }

        public string EncryptionKey { get; set; }
    }
}