﻿using System;
using System.IO;
using System.IO.Compression;

namespace Plywood.IO
{
    public interface IInputOutput {
        Stream CreateStream();

        Stream BaseStream { get; }
    }

    public class Input : IInputOutput {
        private Stream p_BaseStream;

        public Input(Stream stream) {
            //Check the stream
            if (!stream.CanRead) {
                throw new Exception("Invalid stream");
            }

            //Define the globals
            p_BaseStream = stream;
        }

        public virtual int Read(byte[] buffer, int offset, int count) {
            //Read the bytes from the base stream
            return p_BaseStream.Read(buffer, offset, count);
        }
        public virtual int ReadByte() {
            //Read a byte from the base stream
            return p_BaseStream.ReadByte();
        }

        public Stream BaseStream { get { return p_BaseStream; } }

        public Stream CreateStream()  {
            //Let IOStream handle this
            return new IOStream(this, null, p_BaseStream);
        }
    }
    public class Output : IInputOutput {
        private Stream p_BaseStream;
        private long p_Length;

        public Output(Stream stream) {
            //Check the stream
            if (!stream.CanWrite) {
                throw new Exception("Invalid stream");
            }

            //Define the globals
            p_BaseStream = stream;
        }

        public virtual void Write(byte[] buffer, int offset, int count) {
            //Write the bytes to the base stream
            p_BaseStream.Write(buffer, offset, count);

            //Add the count to the global length
            p_Length += count;
        }
        public virtual void WriteByte(byte value) {
            //Write the byte to the base stream
            p_BaseStream.WriteByte(value);

            //Incriment the length
            p_Length++;
        }

        public Stream BaseStream { get { return p_BaseStream; } }
        public Stream CreateStream() {
            //Let IOStream handle this
            return new IOStream(null, this, p_BaseStream);
        }

        public long Length { get { 
            //Get the compressed size?
            if (p_BaseStream is GZipStream) { 
                //Return the actual compression size
                Stream b = ((GZipStream)p_BaseStream).BaseStream;
                if (!(b is IOStream) || ((IOStream)b).Output == null) { return p_Length; }
                return ((IOStream)((GZipStream)p_BaseStream).BaseStream).Output.Length;
            }

            //Just return our length
            return p_Length;
        } }

        public static Output CreateCompression(Stream target) {
            //Create the target write stream
            Stream baseStr = new IOStream(null, new Output(target), target);

            //Create the compression stream
            Stream compression = new GZipStream(baseStr, CompressionMode.Compress);

            //Return the output
            return new Output(compression);
        }
    }
    
    public class IOStream : Stream {
        private Stream p_Stream;
        private Input p_Read;
        private Output p_Write;

        public IOStream(Input read, Output write, Stream stream) {
            //Check the stream
            if ((!stream.CanWrite && !stream.CanRead) ||
               (read == null && write == null)) {
                throw new Exception("Invalid IO");
            }

            //Define the globals
            p_Stream = stream;
            p_Read = read;
            p_Write = write;
        }

        public override void Flush() {
            throw new NotSupportedException();
        }
        public override bool CanRead { get { return p_Read != null; } }
        public override bool CanWrite { get { return p_Write != null; } }
        public override bool CanSeek { get { return p_Stream.CanSeek; } }
        public override void Write(byte[] buffer, int offset, int count) {
            //Can we write?
            if (!CanWrite) {
                throw new NotSupportedException();
            }

            //Write
            p_Write.Write(buffer, offset, count);
        }
        public override void WriteByte(byte value) {
            //Can we write?
            if (!CanWrite) {
                throw new NotSupportedException();
            }

            //Write the byte
            p_Write.WriteByte(value);
        }
        public override int Read(byte[] buffer, int offset, int count) {
            //Can we read?
            if (!CanRead) {
                throw new NotSupportedException();
            }

            //Read
            return p_Read.Read(buffer, offset, count);
        }
        public override int ReadByte() {
            //Are we allowed to read?
            if (!CanRead) {
                throw new NotSupportedException();
            }

            //Read 1 byte
            return p_Stream.ReadByte();
        }
        public override void SetLength(long value) {
            //Set the stream length
            p_Stream.SetLength(value);
        }
        public override long Seek(long offset, SeekOrigin origin) {
            //Seek the base stream
            return p_Stream.Seek(offset, origin);
        }
        
        public override long Length { get { return p_Stream.Length; } }
        public override long Position { get { return p_Stream.Position; } set { p_Stream.Position = value; } }

        public Input Input { get { return p_Read; } }
        public Output Output { get { return p_Write; } }
    }
}