﻿using System;
using System.IO;
using System.Reflection;

namespace Plywood {
    public class RunState {
        private Interpreter p_Script;
        private Stream p_Stream;
        internal Stream p_PrintStreamBuffer;
        internal string p_PrintStringFile;
        
        public event OnConsoleOperationHandler OnConsoleOperation;
        public event OnMethodInvokeHandler OnMethodInvoke;
        
        public RunState(Interpreter script, Stream stream) {
            //Define the globals
            Functions = new GenericMethods(this);
            p_Script = script;
            p_Stream = stream;
            MainStream = stream;
            p_PrintStreamBuffer = FileSystem.CreateTempStream(out p_PrintStringFile);
        }

        internal GenericMethods Functions;
        internal IDefinition CurrentDefinition;
        internal void invokeMethod(string name, IMethod method, ref bool allow) {
            //Invoke the event
            if (OnMethodInvoke != null) {
                OnMethodInvoke(name, method, ref allow);
                return;
            }

            //Just do the default acton
            allow = true;
        }

        public Stream MainStream { get; set; }
        public Stream Stream { get { return p_Stream; } }
        public Interpreter Script { get { return p_Script; } }
        public bool IsElevated { get; set; }
        public Stream ScreenBuffer { get { return p_PrintStreamBuffer; } set { p_PrintStreamBuffer = value; } }
        public string ScreenBufferFilename { get { return p_PrintStringFile; } set { p_PrintStringFile = value; } }
        public bool ApplyDefinitionGlobal { get; set; }

        public static int PrimaryTerminal { get; set; }

        public void TerminalOperation(TerminalMessageType type, object arg) { 
            //Let the local method handle this
            doConsoleOp(type, arg, true);
        }
        public object TerminalOperation(TerminalMessageType type) {
            //Let the local method handle this
            return doConsoleOp(type, null, false);
        }
        
        internal object doConsoleOp(TerminalMessageType type, object arg, bool set) {
            //Check
            if ((type == TerminalMessageType.Write ||
                type == TerminalMessageType.WriteLine ||
                type == TerminalMessageType.ReadKey ||
                type == TerminalMessageType.ReadLine ||
                type == TerminalMessageType.MaximumWindowHeight ||
                type == TerminalMessageType.MaximumWindowWidth ||
                type == TerminalMessageType.ClearScreen) && set) {
                throw new Exception("Invalid function");
            }

            //Update primary terminal?
            if (PrimaryTerminal >= p_Script.p_Terminals.Length || PrimaryTerminal < 0) {
                PrimaryTerminal = 0;
            }

            //Get?
            if (type == TerminalMessageType.ReadKey || type == TerminalMessageType.ReadLine) { 
                //Get from the primary terminal
                return p_Script.p_Terminals[PrimaryTerminal].MessageSent(
                                this, type, arg, false);
            }

            //Loop through every terminal
            foreach (ITerminal t in p_Script.p_Terminals) { 
                //Enabled?
                if (!t.Enabled) { continue; }
                //Send the message
                object r = null;
                try { r = t.MessageSent(this, type, arg, set); }
                catch { }
                //if (!set) { return r; }
            }
            return null;
        }
       
        public delegate void OnMethodInvokeHandler(string name, IMethod method, ref bool allow);
        public delegate object OnConsoleOperationHandler(TerminalMessageType type, object arg, bool set, ref bool handled);
        
        public void Stop() {
            //Push the stream location to the end of the stream
            Stream.Seek(0, SeekOrigin.End);
        }


        public enum LoopFunctionType { 
            Continue = 0,
            Break = 1
        }
    }

}