﻿using System;
using System.IO;
using System.Drawing;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Plywood {
    public class TerminalWindow : Control, ITerminal {
        private Point[] p_UpdateList = new Point[0];
        private charObj[,] p_Matrix;
        private Size p_MatrixSize = new Size(70, 20);
        private Font p_Font;
        private Size p_CellSize;
        private bool p_ClearAll = true;
        private Point p_CursorPosition = new Point(0, 0);
        private Color p_ForeColor = Color.LightGray, p_BackColor = Color.Black;
        private ConsoleColor p_ConsoleForeColor = ConsoleColor.Gray, p_ConsoleBackColor = ConsoleColor.Black;
        private ulong p_RenderTime = 0;
        private int p_CursorAlpha = 0, p_CursorAlphaVel = 20;
        private int p_CursorHeight = 3;
        private string[] p_ReadCache = new string[0];
        private int p_CacheSelect = 0;
        private bool p_HasRendered = false;
        private bool p_ShowCursor = true;
        private bool p_CursorVisible = true;

        private charObj[,] createMatrix() { 
            //Create the matrix
            return new charObj[p_MatrixSize.Width, p_MatrixSize.Height];
        }

        public TerminalWindow() { 
            //Define the font
            p_Font = new Font("Consolas", 8, FontStyle.Bold);
            p_CellSize = new Size((int)p_Font.SizeInPoints, (int)p_Font.GetHeight());
            
            //set the size
            Size = new Size(p_MatrixSize.Width * p_CellSize.Width, p_MatrixSize.Height * p_CellSize.Height);

            //Define the matrix
            p_Matrix = createMatrix();
            for (int x = 0; x < p_MatrixSize.Width; x++) {
                for (int y = 0; y < p_MatrixSize.Height; y++) {
                    p_Matrix[x, y] = new charObj() {
                        Character = ' ',
                        BackColor = p_BackColor,
                        ForeColor = p_ForeColor
                    };
                }
            }

            //Define the heartbeat for the cursor update
            Timer curUpd = new Timer() {
                Interval = 50, 
                Enabled = true
            };
            
            curUpd.Tick += delegate(object sender, EventArgs e) {
                //Render
                engineDraw(!p_HasRendered);
                p_HasRendered = true;

                //Update cursor alpha
                p_CursorAlpha += p_CursorAlphaVel;
                if (p_CursorAlpha >= 100 || p_CursorAlpha <= 0) {
                    p_CursorAlphaVel = -p_CursorAlphaVel;
                }

                //Update?
                if (!p_CursorVisible) {
                    p_CursorAlpha = 0;
                }
            };
                        
            //Set globals
            Enabled = true;
            Text = System.Diagnostics.Process.GetCurrentProcess().Modules[0].FileName;
            PreviewKeyDown += delegate(object sender, PreviewKeyDownEventArgs e) {
                e.IsInputKey = true;
            };
        }
        
        private void engineDraw() { engineDraw(p_ClearAll); }
        private void engineDraw(bool clearAll) { 
            //Attempt to draw
            try { engineDrawCore(clearAll); }
            catch { }
        }
        private void engineDrawCore(bool clearAll) { 
            //Update cursor
            if (p_CursorPosition.X >= p_MatrixSize.Width) {
                p_CursorPosition = new Point(
                        0, p_CursorPosition.Y + 1);

            }
            if (p_CursorPosition.Y >= p_MatrixSize.Height) {
                p_CursorPosition = new Point(
                        0, p_MatrixSize.Height - 1);
                scroll();
            }

            //Only render the cursor?
            if (!clearAll && p_UpdateList.Length == 0) {
                if (!p_ShowCursor) { return; }
                renderCellBackground(CreateGraphics(), p_CursorPosition.X, p_CursorPosition.Y);
                CreateGraphics().FillRectangle(
                        new SolidBrush(Color.FromArgb(Convert.ToInt32(p_CursorAlpha * 1.0f / 100 * 255), invertColor(
                                p_Matrix[p_CursorPosition.X, p_CursorPosition.Y].BackColor
                            ))),
                        new Rectangle(p_CursorPosition.X * p_CellSize.Width, (p_CursorPosition.Y * p_CellSize.Height) + p_CellSize.Height - p_CursorHeight,
                                      p_CellSize.Width, p_CursorHeight));
                return; 
            }

            //Define the bitmap buffer
            Bitmap bitmapBuffer = new Bitmap((p_CellSize.Width * p_MatrixSize.Width) + 16,
                                             (p_CellSize.Height * p_MatrixSize.Height) + 38);
            Graphics buffer = Graphics.FromImage(bitmapBuffer);
            buffer.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;
            
            //Form not of this size?
            if (Width != bitmapBuffer.Width || Height != bitmapBuffer.Height) {
                try { Size = bitmapBuffer.Size; }
                catch { }
            }

            //Render what we need to
            foreach (Point p in p_UpdateList) {
                renderChar(buffer, p.X, p.Y);
            }

            //Render every cell?
            if (clearAll) {
                for (int y = 0; y < p_MatrixSize.Height; y++) {
                    for (int x = 0; x < p_MatrixSize.Width; x++) {
                        renderChar(buffer, x, y);
                    }
                }
            }

            //Clear render list
            p_UpdateList = new Point[0];
            p_ClearAll = false;
            if (p_RenderTime < 3) { p_ClearAll = true; }

            //Draw it
            p_RenderTime++;
            CreateGraphics().DrawImage(bitmapBuffer, new Point(0, 0));

            //Clean up
            bitmapBuffer.Dispose();
            buffer.Dispose();
            bitmapBuffer = null;
            buffer = null;
        }

        private Color changeColorLight(Color color, int light) { 
            //Change the light
            int r = Convert.ToInt32(color.R * 1.0f / 255 * (light * 1.0f / 100 * 255));
            int g = Convert.ToInt32(color.G * 1.0f / 255 * (light * 1.0f / 100 * 255));
            int b = Convert.ToInt32(color.B * 1.0f / 255 * (light * 1.0f / 100 * 255));
            return Color.FromArgb(color.A, r, g, b);
        }
        private void renderChar(Graphics buffer, int x, int y) {
            //Define it's bounds
            Rectangle b = new Rectangle(x * p_CellSize.Width, y * p_CellSize.Height, p_CellSize.Width, p_CellSize.Height);

            //Define the character brush
            LinearGradientBrush cb = new LinearGradientBrush(
                    new Point(b.X, b.Y),
                    new Point(b.X, b.Y + b.Height),
                    p_Matrix[x, y].ForeColor,
                    changeColorLight(p_Matrix[x, y].ForeColor, 40));
           
            //Render the background
            renderCellBackground(buffer, x, y);

            //Draw it
            buffer.DrawString(p_Matrix[x, y].Character.ToString(), p_Font, cb, b.Location);
        }
        private void scroll() { 
            //Define the new matrix
            charObj[,] buffer = createMatrix();

            //Offset by 1 
            for (int y = 1; y < p_MatrixSize.Height; y++) {
                for (int x = 0; x < p_MatrixSize.Width; x++) {
                    buffer[x, y - 1] = new charObj { 
                        Character = p_Matrix[x, y].Character,
                        BackColor = p_BackColor,
                        ForeColor = p_ForeColor
                    };
                }            
            }

            //Create a new row
            for (int x = 0; x < p_MatrixSize.Width; x++) {
                buffer[x, p_MatrixSize.Height - 1] = new charObj() { Character = ' ', BackColor = p_BackColor, ForeColor = p_ForeColor };
            }
            
            //Clean up
            p_Matrix = null;
            p_UpdateList = new Point[0];
            p_RenderTime = 0;
            p_Matrix = buffer;
            p_ClearAll = true;
            engineDraw(true);
        }
        private void changeMatrixSize(Size size) { 
            //Define the new matrix 
            charObj[,] buffer = new charObj[size.Width, size.Height];

            //Fill in the matrix
            for (int x = 0; x < size.Width; x++) {
                for (int y = 0; y < size.Height; y++) {
                    buffer[x, y] = new charObj { BackColor = p_BackColor, ForeColor = p_ForeColor, Character = ' '};
                    if (x >= p_MatrixSize.Width || y >= p_MatrixSize.Height) { continue; }
                    buffer[x, y] = p_Matrix[x, y];
                }
            }

            //Resize (invalidate it)
            Invoke(new MethodInvoker(delegate { Size = new Size(40, 40); }));

            //Clean up
            p_Matrix = null;
            p_Matrix = buffer;
            p_MatrixSize = size;
            p_ClearAll = true;
            engineDraw(true);
        }
        private void renderCellBackground(Graphics buffer, int x, int y) {
            //Define it's bounds
            Rectangle b = new Rectangle(x * p_CellSize.Width, y * p_CellSize.Height, p_CellSize.Width, p_CellSize.Height);

            //Define the background brush
            LinearGradientBrush bb = new LinearGradientBrush(
                    new Point(0, 0),
                    new Point(0, Height),
                    p_Matrix[x, y].BackColor,
                    changeColorLight(p_Matrix[x, y].BackColor, 80));

            //Draw it
            buffer.FillRectangle(bb, b);
        }

        public object MessageSent(RunState state, TerminalMessageType type, object arg, bool set) {
            //Write line?
            if (type == TerminalMessageType.WriteLine) {
                return MessageSent(state, TerminalMessageType.Write, arg.ToString() + "\n", set);
            }
            
            //Write?
            if (type == TerminalMessageType.Write) {
                foreach (char c in arg.ToString()) {
                    if (c == '\b') { continue; }
                    if (c == '\n' || c == '\r') {
                        Array.Resize(ref p_UpdateList, p_UpdateList.Length + 1);
                        p_UpdateList[p_UpdateList.Length - 1] = p_CursorPosition;
                        p_CursorPosition = new Point(0, p_CursorPosition.Y + 1);
                        if (p_CursorPosition.Y >= p_MatrixSize.Height) {
                            p_CursorPosition = new Point(0, p_CursorPosition.Y - 1);
                            scroll();
                        }
                        continue; 
                    }
                    if (p_CursorPosition.X >= p_MatrixSize.Width) {
                        p_CursorPosition = new Point(0, p_CursorPosition.Y + 1);
                    }
                    if (p_CursorPosition.Y >= p_MatrixSize.Height) {
                        p_CursorPosition = new Point(0, p_MatrixSize.Height - 1);
                        scroll();
                    }
                    p_Matrix[p_CursorPosition.X, p_CursorPosition.Y].Character = c;
                    p_Matrix[p_CursorPosition.X, p_CursorPosition.Y].BackColor = p_BackColor;
                    p_Matrix[p_CursorPosition.X, p_CursorPosition.Y].ForeColor = p_ForeColor;
                    Array.Resize(ref p_UpdateList, p_UpdateList.Length + 1);
                    p_UpdateList[p_UpdateList.Length - 1] = p_CursorPosition;
                    p_CursorPosition = new Point(p_CursorPosition.X + 1, p_CursorPosition.Y);
                }
                engineDraw();              
                return null;
            }
            
            //Read key?
            if (type == TerminalMessageType.ReadKey) {
                char rKey = (char)0;
                KeyPress += delegate(object sender, KeyPressEventArgs e) { rKey = e.KeyChar; };
                while (rKey == (char)0) {
                    System.Threading.Thread.Sleep(1);
                }
                MessageSent(state, TerminalMessageType.Write, rKey.ToString(), false);
                return rKey.ToString();
            }
            //Read line?
            if (type == TerminalMessageType.ReadLine) {
                string read = "";
                Point[] positions = new Point[0];
                Point cursorPosOld = p_CursorPosition;
                while (true) {
                    char rChar = (char)0;
                    Keys key = Keys.A;

                    //Define the key handlers
                    KeyEventHandler keyDown = delegate(object sender, KeyEventArgs e) {
                        key = e.KeyCode;
                    };
                    KeyPressEventHandler keyPress = delegate(object sender, KeyPressEventArgs e) {
                        rChar = e.KeyChar;                         
                    };
                    KeyDown += keyDown;
                    KeyPress += keyPress;

                    //Read the keys
                    while (rChar == (char)0) {
                        if ((key == Keys.Up || key == Keys.Down) && p_ReadCache.Length != 0) { break; }
                        System.Threading.Thread.Sleep(1);
                    }

                    //Clean up the handlers
                    KeyDown -= keyDown;
                    KeyPress -= keyPress;

                    //Push value from cache?
                    if ((key == Keys.Up || key == Keys.Down) && rChar == '\0') {                       
                        //Select the next cache item
                        int cacheIndex = p_CacheSelect;
                        if (key == Keys.Up && p_CacheSelect >= p_ReadCache.Length) { p_CacheSelect = 0; cacheIndex = 0; }
                        else if (key == Keys.Up) { p_CacheSelect++; }
                        if (key == Keys.Down && p_CacheSelect == 0) { p_CacheSelect = p_ReadCache.Length - 1; cacheIndex = p_CacheSelect; }
                        else if (key == Keys.Down) { p_CacheSelect--; }
                        
                        //Display the cache item
                        read = p_ReadCache[cacheIndex];
                        foreach (Point p in positions) { 
                            p_Matrix[p.X, p.Y].Character = ' ';
                            Array.Resize(ref p_UpdateList, p_UpdateList.Length + 1);
                            p_UpdateList[p_UpdateList.Length - 1] = p;
                        }

                        positions = new Point[read.Length];
                        p_CursorPosition = cursorPosOld;
                        for (int c = 0; c < read.Length; c++) {
                            MessageSent(state, TerminalMessageType.Write, read[c].ToString(), false);
                            positions[c] = p_CursorPosition;
                        }

                        continue;
                    }

                    //Write
                    MessageSent(state, TerminalMessageType.Write, rChar.ToString(), false);
                    Array.Resize(ref positions, positions.Length + 1);
                    positions[positions.Length - 1] = p_CursorPosition;

                    if (rChar == '\b') {
                        Point curOld = p_CursorPosition;
                        if (read.Length == 0) { continue; }
                        if (p_CursorPosition.X == 0) { p_CursorPosition = new Point(p_MatrixSize.Width - 1, p_CursorPosition.Y - 1); }
                        else { p_CursorPosition = new Point(p_CursorPosition.X - 1, p_CursorPosition.Y); }
                        read = read.Substring(0, read.Length - 1);
                        p_Matrix[p_CursorPosition.X, p_CursorPosition.Y].Character = ' ';
                        Array.Resize(ref p_UpdateList, p_UpdateList.Length + 2);
                        p_UpdateList[p_UpdateList.Length - 1] = p_CursorPosition;
                        p_UpdateList[p_UpdateList.Length - 2] = curOld;

                        engineDraw();
                        continue;
                    }
                    if (rChar == '\n' || rChar == '\r') { break; }
                    read += rChar;
                }
                //Add to cache if it does not exist
                bool cacheExist = false;
                foreach (string s in p_ReadCache) {
                    if (s == read) { cacheExist = true; break; }
                }
                if (!cacheExist) {
                    Array.Resize(ref p_ReadCache, p_ReadCache.Length + 1);
                    p_ReadCache[p_ReadCache.Length - 1] = read;
                }
                return read;
            }
            //Clear screen?
            if (type == TerminalMessageType.ClearScreen) {
                p_ShowCursor = false;
                for (int x = 0; x < p_MatrixSize.Width; x++) {
                    for (int y = 0; y < p_MatrixSize.Height; y++) {
                        p_Matrix[x, y].ForeColor = p_ForeColor;
                        p_Matrix[x, y].BackColor = p_BackColor;
                        p_Matrix[x, y].Character = ' ';
                    }
                }
                p_ClearAll = true;
                p_CursorPosition = new Point(0, 0);
                p_ShowCursor = true;
                engineDraw(true);
            }
            //Set fore color?
            if (type == TerminalMessageType.ForeColor) {
                if (!set) { return p_ConsoleForeColor; }
                p_ConsoleForeColor = (ConsoleColor)arg;
                p_ForeColor = consoleColorToColor(p_ConsoleForeColor);
            }
            //Set back color?
            else if (type == TerminalMessageType.BackColor) {
                if (!set) { return p_ConsoleBackColor; }
                p_ConsoleBackColor = (ConsoleColor)arg;
                p_BackColor = consoleColorToColor(p_ConsoleBackColor);
            }
            //Set cursor position
            else if (type == TerminalMessageType.CursorX) {
                if (!set) { return p_CursorPosition.X; }
                Array.Resize(ref p_UpdateList, p_UpdateList.Length + 1);
                p_UpdateList[p_UpdateList.Length - 1] = p_CursorPosition;
                p_CursorPosition = new Point((int)arg, p_CursorPosition.Y);                
            }
            else if (type == TerminalMessageType.CursorY) {
                if (!set) { return p_CursorPosition.Y; }
                Array.Resize(ref p_UpdateList, p_UpdateList.Length + 1);
                p_UpdateList[p_UpdateList.Length - 1] = p_CursorPosition;
                p_CursorPosition = new Point(p_CursorPosition.X, (int)arg);
            }
            //Title?
            if (type == TerminalMessageType.Title) {
                if (!set) { return Text; }                    
                Invoke(new MethodInvoker(delegate {
                    Text = (string)arg;
                }));
            }
            //Window size
            if (type == TerminalMessageType.WindowWidth) {
                if (!set) { return p_MatrixSize.Width; }
                changeMatrixSize(new Size((int)arg, p_MatrixSize.Height));
            }
            else if (type == TerminalMessageType.WindowHeight) {
                if (!set) { return p_MatrixSize.Height; }
                changeMatrixSize(new Size(p_MatrixSize.Width, (int)arg));
            }
            //Cursor size
            if (type == TerminalMessageType.CursorSize) {
                if (!set) { return p_CursorHeight; }
                if ((int)arg > p_CellSize.Height) { arg = (int)p_CellSize.Height; }
                p_CursorHeight = (int)arg;
            }
            //Cursor visible?
            if (type == TerminalMessageType.CursorVisible) {
                if (!set) { return p_CursorVisible; }
                p_CursorVisible = Convert.ToBoolean(arg);
            }

            return null;
        }

        public Form CreateForm(bool exitOnClose) { 
            //Define the buffer
            Form frm = new Form();

            //Add this control to the form
            frm.Controls.Add(this);

            //Set the form globals
            frm.ShowIcon = false;
            frm.MaximizeBox = false;
            frm.Text = Text;
            frm.MinimumSize = Size;
            frm.MaximumSize = Size;
            Resize += delegate(object sender, EventArgs e) {
                frm.Size = Size;
                frm.MaximumSize = Size;
                frm.MinimumSize = Size;
            };
            TextChanged += delegate(object sender, EventArgs e) {
                frm.Text = Text;
            };
            frm.FormClosed += delegate(object sender, FormClosedEventArgs e) {
                if (CommandSystem.Terminal == this || exitOnClose) {
                    System.Diagnostics.Process.GetCurrentProcess().Kill();
                }
            };

            //Return the form
            return frm;
        }
        
        protected override void OnPaint(PaintEventArgs e) {
            engineDraw(true);
        }

        public bool Enabled { get; set; }
        public Color BackColor { get { return p_BackColor; } set { p_BackColor = value; } }
        public Color ForeColor { get { return p_ForeColor; } set { p_ForeColor = value; } }


        private Color consoleColorToColor(ConsoleColor color) {
            //Return the appropriate color
            switch (color) {
                case ConsoleColor.Black: return Color.Black;
                case ConsoleColor.Blue: return Color.Blue;
                case ConsoleColor.Cyan: return Color.Cyan;
                case ConsoleColor.DarkBlue: return Color.DarkBlue;
                case ConsoleColor.DarkCyan: return Color.DarkCyan;
                case ConsoleColor.DarkGray: return Color.Gray;
                case ConsoleColor.DarkGreen: return Color.Green;
                case ConsoleColor.DarkMagenta: return Color.Magenta;
                case ConsoleColor.DarkRed: return Color.DarkRed;
                case ConsoleColor.DarkYellow: return Color.OrangeRed;
                case ConsoleColor.Gray: return Color.LightGray;
                case ConsoleColor.Green: return Color.Lime;
                case ConsoleColor.Magenta: return Color.Magenta;
                case ConsoleColor.Red: return Color.Red;
                case ConsoleColor.White: return Color.White;
                case ConsoleColor.Yellow: return Color.Yellow;
            }

            //Return the default color
            return Color.White;
        }
        private Color invertColor(Color clr) { 
            //Invert the color
            return Color.FromArgb(clr.A, 255 - clr.R, 255 - clr.G, 255 - clr.B);
        }

        private struct charObj {
            public char Character;
            public Color BackColor;
            public Color ForeColor;
        }

     
    }
}